from udsp.DS import *
FREE = 0
BUSY = 1
BUSY_UPDATE = 2
CMDTYPES = {  # 不支持3字节长度的cmd
    1: 'B',
    2: 'H',
    4: 'I',
}

class DSP:
    @property
    def cl(self):
        return self.__cl
    def _get_cmd(self):
        return self.__cmd
    def _set_cmd(self, v):
        if self.tx_state != FREE:
            raise Exception("[Error]: You can only change <cmd> before <Encode> and after <TxReset>.")
        self.__cmd = v
    cmd = property(_get_cmd, _set_cmd)

    def __init__(self, txbuf, rxbuf, callback):
        self.cb = callback
        self.__cl = 1  # 保留属性，禁止修改
        self.__cmd_type = CMDTYPES[self.__cl]

        self.__cmd = 0
        self.__cl_cmd = 0xfe  # (2^cl - 1) - 1

        self.Loading(txbuf, rxbuf)

    def Loading(self, *args):
        self.rx_state = FREE
        self.tx_state = FREE
        self.txbuf = args[0]
        self.rxbuf = args[1]

        self.TxReset()
        self.RxReset()


    def RxReset(self):
        self.rx_state = FREE
        self.rds = DS(self.rxbuf)  # 为下次接收做好准备

    def TxReset(self):
        if (self.tx_state == BUSY_UPDATE):
            self.tds.O(self.__cmd_type)
            self.__ChangeCL(self.tds.O('B'))
        self.tx_state = FREE
        self.tds = DS(self.txbuf)  # 为下次发送做好准备

    def Encode(self, fmt, *data):
        """
        编码数据到tx_buffer

        注意，在发送数据前至少应该Encode过一次
        你可以使用.Encode (None)来添加空数据.

        注意，发送完成后记得TxReset
        :param fmt: 适用于struct的fmt 可以为None
        :param data:
        :return:
        """
        if self.tx_state == FREE:
            self.tx_state = BUSY
            self.__LoadTxCmd()
        if fmt is None: return
        self.tds.I(fmt, *data)

    def EncodeCL(self, new_cl):
        """
        修改双方的cmd length

        注意，调用该函数会覆盖txbuf已有的内容.

        注意，发送完成后记得TxReset
        :param new_cl: 新的cl，只能为1 2 4中的一种
        :return:
        """
        self.TxReset()
        self.cmd = self.__cl_cmd
        self.Encode('B', 2)

    def Decode(self, length=None):
        """
        解析rxbuf中的数据
        自动调用回调以及RxReset
        :param length: rxbuf中有多少个byte的数据, 不传参数时全部进行解析
        :return:
        """
        if length is None: length = len(self.rxbuf)
        if not length: return
        while self.rx_state != FREE: ...
        self.rx_state = BUSY
        #debug# print("rxlength:", length)
        self.rds.len = length

        cmd = self.__ParseRxCmd()
        if cmd is None: return

        self.cb(cmd, self.rds)

        self.RxReset()

    def __LoadTxCmd(self):
        self.tds.I(self.__cmd_type, self.cmd)
        if self.cmd == self.__cl_cmd:
            self.tx_state = BUSY_UPDATE

    def __ParseRxCmd(self):
        rtn = self.rds.O(self.__cmd_type)
        if rtn == self.__cl_cmd:
            self.__ChangeCL(self.rds.O('B'))
            return
        return rtn

    def __ChangeCL(self, new_cl):
        if (new_cl == self.__cl): return
        new_type = CMDTYPES.get(new_cl)
        if (new_type is None):
            raise Exception("[Error]: Unsupported CL<:{}>.".format(self.__cl))

        self.__cl = new_cl
        self.__cmd_type = new_type
        self.__cl_cmd = 2 ** new_cl - 2

    @property
    def txlen(self):
        """
        允许通过该属性获取txbuf中有效数据的长度
        :return:
        """
        return self.tds.len

    @property
    def rxlen(self):
        """
        允许通过该属性获取rxbuf中有效数据的长度
        :return:
        """
        return self.rds.len
    def __str__(self):
        return "<DSPState>\n\t<cmd:>{}\n\t<cl:>{}\n\t<TxState:>{}\n\t<RxState:>{}\n".format(self.cmd, self.cl, self.tds, self.rds)
    def __repr__(self):
        return str(self)
    def __len__(self):
        return dict(tx=self.txlen, rx=self.rxlen)

if __name__ == '__main__':
    class Compound:
        def __init__(self, i, d):
            self.i = i
            self.d = d
        def __repr__(self):
            return "Compound: <i: {}; d: {}>".format(self.i, self.d)

    def callback(cmd, ds):
        if cmd == 1:
            print("Get CMD:1, int =", ds.O('I'))
        elif cmd == 2:
            print("Get CMD:2, compound =", Compound(*ds.O('Id')))

    ba1, ba2 = bytearray(60), bytearray(60)
    dsp1, dsp2 = DSP(ba1, ba2, callback), DSP(ba2, ba1, callback)

    print(dsp1, dsp2)

    dsp1.cmd = 1
    dsp1.Encode('I', 114514)
    dsp2.Decode(dsp1.txlen)
    dsp1.TxReset()

    dsp2.cmd = 2
    dsp2.Encode('Id', 114, .514)
    dsp1.Decode(dsp2.txlen)
    dsp2.TxReset()

    dsp1.EncodeCL(2)
    dsp2.Decode(dsp1.txlen)
    dsp1.TxReset()

    print(dsp1, dsp2)





class ARXTD:
    def __init__(self, source, endl, empty=None):
        self.s = source
        self.endl = endl
        self.empty = empty
        
        self.busy = False
        self.buf = []
        self.piece = bytearray()
    
        
    
    def Update(self):
        tmp = self.s()
        if tmp == self.empty: return
        #print(tmp)
        if len(self.piece): # 从上次的中断处继续
            
            end_index = tmp.find(self.endl)
            if end_index != -1:
                self.Push(self.piece + tmp[:end_index])
            else:
                self.piece += tmp
        else:  # 新的操作
            while len(tmp):
                end_index = tmp.find(self.endl)
                if end_index != -1:
                    self.Push(tmp[:end_index])
                    tmp = tmp[end_index + 2:]
                else:
                    self.piece = tmp
                    break
    
    def Push(self, item):
        while self.busy: ...
        self.busy = True
        self.buf += [self.RemoveZeros(item)]
        self.busy = False
        
    def Pop(self, index=0):
        while self.busy: ...
        self.busy = True
        tmp = self.buf.pop(0)
        self.busy = False
        return tmp
    
    def __len__(self):
        return len(self.buf)
        
    @staticmethod
    def RemoveZeros(item):  # 去除开头的若干个0
        i = 0
        while item[i] == 0:
            i += 1
        return item[i:]
    


from udsp.ARXTD import *
from udsp.DSP import *
import serial
import time

class UDSP:  # serial DSP
    def __init__(self, port, cb, baudrate=115200, txsize=100):
        self.alive = True
        self.txbuf = bytearray(txsize)
        
        self.ser = serial.Serial(port, baudrate, timeout=0.5) 
        
        self.arxtd = ARXTD(self.ser.read_all, b'\r\n', b'')
        self.dsp = DSP(self.txbuf, bytearray(1), cb)
        
    def Retribute(self, new_buf):
        self.dsp.rxbuf = new_buf
        self.dsp.RxReset()

    def Cmd(self, v):
        self.dsp.TxReset()
        self.dsp.cmd = v
    
    def Decode(self):
        self.arxtd.Update()
        while len(self.arxtd):
            self.Retribute(self.arxtd.Pop())
            self.dsp.Decode()
    
    def Encode(self, fmt, *data):
        self.dsp.Encode(fmt, *data)
        

if __name__ == "__main__":
	def func(cmd, ds):
		if cmd == 1:
			print(ds.O('B'))
	udsp = UDSP("COM3", func)
	while True:
		time.sleep(1)
		sdsp.Decode()
		#print(tmp)



import struct

class DS:
    def __init__(self, buf, index=0, length=0, left=None):
        self.buf = buf

        if not left: left = len(self.buf) - length - index
        self.Loading(index, length, left)

    def Loading(self, *args):
        self.index = args[0]
        self.len = args[1]
        self.left = args[2]

    @staticmethod
    def CalcRawSize(fmt):
        i, rec, _len, _size, _child = 0, 0, len(fmt), 0, ''
        while i < _len:
            if fmt[i].isalpha():
                rec = i
                i += 1
                while _len - i and not fmt[i].isalpha():
                    i += 1
                i -= 1
            else:
                i += 1
                continue

            i += 1

            #debug# print(':', rec, i, fmt[rec:i])
            a = struct.calcsize(fmt[rec:i])
            _size += a
            #debug# print("raw add:", a)
        return _size

    def O(self, fmt):
        fmt = '<' + fmt
        size = self.CalcRawSize(fmt)
        #debug# print("calc size ", size)
        if (self.len < size): raise Exception("[Error]: Ran out <length> of stream.\n\tfmt:%s\n\tlen = %d\n\tindex = %d\n\tleft = %d\n\tdata:%s\n"%(str(fmt), self.len, self.index, self.left, self.buf))
        rtn = struct.unpack_from(fmt, self.buf, self.index)
        self.index += size
        self.len -= size
        return rtn if len(rtn) != 1 else rtn[0]


    def I(self, fmt, *data):
        fmt = '<' + fmt
        size = self.CalcRawSize(fmt)
        #debug# print("add len:", size)
        if (self.left < size): raise Exception("[Error]: fill over <left> of stream.\n\tfmt:%s\n\tlen = %d\n\tindex = %d\n\tleft = %d\n\tdata:%s\n"%(str(fmt), self.len, self.index, self.left, self.buf))
        struct.pack_into(fmt, self.buf, self.index + self.len, *data)
        self.len += size
        #debug# print("len ", self.len)
        self.left -= size

    def __str__(self):
        return "<content:>{}...; <index:>{}; <len:>{}; <left:>{}".format(self.buf[:10], self.index, self.len, self.left)
    def __repr__(self):
        return str(self)
    def __len__(self):
        return self.len

if __name__ == '__main__':
    ba = bytearray(60)
    ds = DS(ba)
    ds.I('I', 4096)
    print(ba)



#!/usr/bin/env python
# coding:utf-8

from setuptools import find_packages, setup

setup(
name='udsp',
version='0.1.1',
description='Uart Data Stream Processor (Whether using Uart is not important at all. it can also support other commulication protocal.). It is used to format C data into strings and easily manage send and receive buffers in python and micropython.(We used it with uart in Electronic Design Competition.)',
author="Eagle'sBaby",
author_email='2229066748@qq.com',
maintainer="Eagle'sBaby",
maintainer_email='2229066748@qq.com',
packages=find_packages(),
platforms=["all"],
license='Apache Licence 2.0',
classifiers=[
'Programming Language :: Python',
'Programming Language :: Python :: 3',
],
keywords = ['pymouse', 'PyHook3'],
python_requires='>=3', 
)


from udsp.DSP import *
from udsp.UDSP import *

"""
数据流处理器
Data Stream Processor

这个DSP能管理一个发送buffer和一个接收buffer，使用struct进行结构化。

def call_back(cmd, ds):
	if cmd == 1:
		print("收到了1号命令，1号命令的数据为:", ds.O('B'))  # 1号命令传输什么样的数据 需要双方事先约定

txbuf, rxbuf = bytearray(40), bytearray(40)

tx, rx = DSP(txbuf, rxbuf, None), DSP(rxbuf, txbuf, call_back)

tx.Cmd(1)  # 设置命令号(默认命令号的内存占用1个字节，即不能超过256。由于协议要求， 0和0xff..是不能用的.)
tx.Encode('B', 155);

rx.Decode()
tx.TxReset()  # 发送完成后必须清除发送信息

# --------------------------------------

tx.EncodeCL(2)  # 修改命令号的内存占用为2，即命令号不能超过65536。 由于协议要求， 0和0xffff..是不能用的.
# 只能设置为1 2 4中的一种

rx.Decode()  # 对方收到后会自动修改命令号的内存占用，而不会进入call_back.
tx.TxReset()  # 发送完成后必须清除发送信息

#// -------- ####################################################### ----------------------

This DSP can manage a transmit buffer and a receive buffer, and use struct for structure.

def call_back(cmd, ds):
	if cmd == 1:
		print("Command 1 received. The data of command 1 is:", ds.O('B'))  # What kind of data should be transmitted by order 1? Both parties need to agree in advance

txbuf, rxbuf = bytearray(40), bytearray(40)

tx, rx = DSP(txbuf, rxbuf, None), DSP(rxbuf, txbuf, call_back)

tx.Cmd(1)  # Set the command number (the memory of the default command number occupies 1 byte, that is, it cannot exceed 256. Due to protocol requirements, 0 and 0xff.. are unavailable.)
tx.Encode('B', 155);

rx.Decode()
tx.TxReset()  # After sending, the sending information must be cleared
# --------------------------------------

tx.EncodeCL(2)  # The memory consumption of modifying the command number is 2, that is, the command number cannot exceed 65536. 0 and 0xFFFF.. are not available due to protocol requirements
# Can only be set to one of 1 2 4

rx.Decode()  # After receiving the command, the other party will automatically modify the memory occupation of the command number without entering call_back.
tx.TxReset()  # After sending, the sending information must be cleared


"""




# Example
"""
from dsp import *
class Compound:
	def __init__(self, i, d):
		self.i = i
		self.d = d
	def __repr__(self):
		return "Compound: <i: {}; d: {}>".format(self.i, self.d)

def callback(cmd, ds):
	if cmd == 1:
		print("Get CMD:1, int =", ds.O('I'))
	elif cmd == 2:
		print("Get CMD:2, compound =", Compound(*ds.O('Id')))

ba1, ba2 = bytearray(60), bytearray(60)
dsp1, dsp2 = DSP(ba1, ba2, callback), DSP(ba2, ba1, callback)

print(dsp1, dsp2)

dsp1.cmd = 1
dsp1.Encode('I', 114514)
dsp2.Decode(dsp1.txlen)
dsp1.TxReset()

dsp2.cmd = 2
dsp2.Encode('Id', 114, .514)
dsp1.Decode(dsp2.txlen)
dsp2.TxReset()

dsp1.EncodeCL(2)
dsp2.Decode(dsp1.txlen)
dsp1.TxReset()

print(dsp1, dsp2)
"""

#Example with uart(series)
"""
def func(cmd, ds):
	if cmd == 1:
		print(ds.O('B'))
udsp = UDSP("COM3", func)
while True:
	time.sleep(1)
	sdsp.Decode()
	#print(tmp)
"""

